#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from humanize import naturaltime

from paasta_tools.cli.cmds.mark_for_deployment import mark_for_deployment
from paasta_tools.cli.utils import extract_tags
from paasta_tools.cli.utils import figure_out_service_name
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import list_deploy_groups
from paasta_tools.cli.utils import validate_full_git_sha
from paasta_tools.cli.utils import validate_given_deploy_groups
from paasta_tools.deployment_utils import get_currently_deployed_sha
from paasta_tools.remote_git import list_remote_refs
from paasta_tools.utils import _log_audit
from paasta_tools.utils import datetime_from_utc_to_local
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import format_table
from paasta_tools.utils import get_git_url
from paasta_tools.utils import list_services
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import parse_timestamp
from paasta_tools.utils import RollbackTypes


def add_subparser(subparsers):
    list_parser = subparsers.add_parser(
        "rollback",
        help="Rollback a docker image to a previous deploy",
        description=(
            "'paasta rollback' is a human-friendly tool for marking a particular "
            "docker image for deployment, which invokes a bounce. While the command "
            "is called 'rollback', it can be used to roll forward or back, as long "
            "as there is a docker image available for the input git SHA."
        ),
        epilog=(
            "This rollback command uses the Git control plane, which requires network "
            "connectivity as well as authorization to the git repo."
        ),
    )
    list_parser.add_argument(
        "-k",
        "--commit",
        help="Git SHA to mark for rollback. "
        "A commit to rollback to is required for paasta rollback to run. However if one is not provided, "
        "paasta rollback will instead output a list of valid git shas to rollback to.",
        required=False,
        type=validate_full_git_sha,
    ).completer = lazy_choices_completer(list_previously_deployed_shas)
    list_parser.add_argument(
        "-l",
        "--deploy-groups",
        help="Mark one or more deploy groups to roll back (e.g. "
        '"all.main", "all.main,all.canary"). If no deploy groups specified,'
        " all deploy groups for that service are rolled back",
        default="",
        required=False,
    ).completer = lazy_choices_completer(list_deploy_groups)
    list_parser.add_argument(
        "-s", "--service", help='Name of the service to rollback (e.g. "service1")'
    ).completer = lazy_choices_completer(list_services)
    list_parser.add_argument(
        "-y",
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    list_parser.add_argument(
        "-f",
        "--force",
        help=("Do not check if Git SHA was marked for deployment previously."),
        action="store_true",
    )
    list_parser.set_defaults(command=paasta_rollback)


def list_previously_deployed_shas(parsed_args, **kwargs):
    service = parsed_args.service
    soa_dir = parsed_args.soa_dir
    deploy_groups = {
        deploy_group
        for deploy_group in parsed_args.deploy_groups.split(",")
        if deploy_group
    }
    return (sha for sha in get_git_shas_for_service(service, deploy_groups, soa_dir))


def get_git_shas_for_service(service, deploy_groups, soa_dir):
    """Returns a dictionary of 2-tuples of the form (timestamp, deploy_group) for each deploy sha"""
    if service is None:
        return []
    git_url = get_git_url(service=service, soa_dir=soa_dir)
    all_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    deploy_groups, _ = validate_given_deploy_groups(all_deploy_groups, deploy_groups)
    previously_deployed_shas = {}
    for ref, sha in list_remote_refs(git_url).items():
        regex_match = extract_tags(ref)
        try:
            deploy_group = regex_match["deploy_group"]
            tstamp = regex_match["tstamp"]
        except KeyError:
            pass
        else:
            # Now we filter and dedup by picking the most recent sha for a deploy group
            # Note that all strings are greater than ''
            if deploy_group in deploy_groups:
                tstamp_so_far = previously_deployed_shas.get(sha, ("all", ""))[1]
                if tstamp > tstamp_so_far:
                    previously_deployed_shas[sha] = (tstamp, deploy_group)
    return previously_deployed_shas


def list_previous_commits(service, deploy_groups, any_given_deploy_groups, git_shas):
    def format_timestamp(tstamp):
        return naturaltime(datetime_from_utc_to_local(parse_timestamp(tstamp)))

    paasta_print("Below is a list of recent commits:")
    git_shas = sorted(git_shas.items(), key=lambda x: x[1], reverse=True)[:10]
    rows = [("Timestamp -- UTC", "Human time", "deploy_group", "Git SHA")]
    for sha, (timestamp, deploy_group) in git_shas:
        rows.extend([(timestamp, format_timestamp(timestamp), deploy_group, sha)])
    for line in format_table(rows):
        paasta_print(line)
    if len(git_shas) >= 2:
        sha, (timestamp, deploy_group) = git_shas[1]
        deploy_groups_arg_line = (
            "-l %s " % ",".join(deploy_groups) if any_given_deploy_groups else ""
        )
        paasta_print(
            "\nFor example, to use the second to last commit from {} used on {}, run:".format(
                format_timestamp(timestamp), PaastaColors.bold(deploy_group)
            )
        )
        paasta_print(
            PaastaColors.bold(
                f"    paasta rollback -s {service} {deploy_groups_arg_line}-k {sha}"
            )
        )


def paasta_rollback(args):
    """Call mark_for_deployment with rollback parameters
    :param args: contains all the arguments passed onto the script: service,
    deploy groups and sha. These arguments will be verified and passed onto
    mark_for_deployment.
    """
    soa_dir = args.soa_dir
    service = figure_out_service_name(args, soa_dir)

    git_url = get_git_url(service, soa_dir)
    given_deploy_groups = {
        deploy_group for deploy_group in args.deploy_groups.split(",") if deploy_group
    }

    all_deploy_groups = list_deploy_groups(service=service, soa_dir=soa_dir)
    deploy_groups, invalid = validate_given_deploy_groups(
        all_deploy_groups, given_deploy_groups
    )

    if len(invalid) > 0:
        paasta_print(
            PaastaColors.yellow(
                "These deploy groups are not valid and will be skipped: %s.\n"
                % (",").join(invalid)
            )
        )

    if len(deploy_groups) == 0:
        paasta_print(
            PaastaColors.red(
                "ERROR: No valid deploy groups specified for %s.\n" % (service)
            )
        )
        return 1

    git_shas = get_git_shas_for_service(service, deploy_groups, soa_dir)
    commit = args.commit
    if not commit:
        paasta_print("Please specify a commit to mark for rollback (-k, --commit).")
        list_previous_commits(
            service, deploy_groups, bool(given_deploy_groups), git_shas
        )
        return 1
    elif commit not in git_shas and not args.force:
        paasta_print(PaastaColors.red("This Git SHA has never been deployed before."))
        paasta_print(
            "Please double check it or use --force to skip this verification.\n"
        )
        list_previous_commits(
            service, deploy_groups, bool(given_deploy_groups), git_shas
        )
        return 1

    returncode = 0

    for deploy_group in deploy_groups:
        rolled_back_from = get_currently_deployed_sha(service, deploy_group)
        returncode |= mark_for_deployment(
            git_url=git_url, service=service, deploy_group=deploy_group, commit=commit
        )

        # we could also gate this by the return code from m-f-d, but we probably care more about someone wanting to
        # rollback than we care about if the underlying machinery was successfully able to complete the request
        if rolled_back_from != commit:
            audit_action_details = {
                "rolled_back_from": rolled_back_from,
                "rolled_back_to": commit,
                "rollback_type": RollbackTypes.USER_INITIATED_ROLLBACK.value,
                "deploy_group": deploy_group,
            }
            _log_audit(
                action="rollback", action_details=audit_action_details, service=service
            )

    return returncode



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from paasta_tools.metrics.metastatus_lib import assert_quorum_size
from paasta_tools.utils import paasta_print


def check_mesos_quorum():
    result = assert_quorum_size()
    if result.healthy:
        paasta_print("OK: " + result.message)
        sys.exit(0)
    else:
        paasta_print(result.message)
        sys.exit(2)


if __name__ == "__main__":
    check_mesos_quorum()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import sys
from collections import defaultdict

from bravado.exception import HTTPError

from paasta_tools.api.client import get_paasta_api_client
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print


def parse_capacity_check_options():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "type", choices=["cpus", "mem", "disk"], type=str, help="The resource to check."
    )
    parser.add_argument(
        "--warn",
        "-w",
        dest="warn",
        type=float,
        default=80,
        help="Level to emit a warning status.",
    )
    parser.add_argument(
        "--crit",
        "-c",
        dest="crit",
        type=float,
        default=90,
        help="Level to emit a critical status.",
    )
    parser.add_argument(
        "--overrides",
        dest="overrides",
        type=str,
        help="json file of per-attribute overrides.\n"
        "In the format [{groupings: {attribute: value, ...}, warn: {cpus: num, disk: num, mem: num}, "
        "crit: {cpus: num, disk: num, mem: num}}, ...]",
    )
    parser.add_argument(
        "--cluster",
        dest="cluster",
        type=str,
        help="Cluster to check. Defaults to looking for the current cluster.",
    )
    parser.add_argument(
        "--attributes",
        dest="attributes",
        type=str,
        default="pool",
        help="Comma separated list of attributes to check.\n"
        "Checks combinations of attributes",
    )
    options = parser.parse_args()

    return options


def calc_percent_usage(resource_item, value_to_check):
    values = resource_item[value_to_check]
    if values["total"] == 0:
        return 0
    return 100 * (values["used"] / values["total"])


def error_message(failures, level, cluster, value_to_check):
    result = f"{level} cluster {cluster} {value_to_check} usage:\n"
    results = []
    for f in failures:
        attrs = ", ".join(["{}: {}".format(e["attr"], e["value"]) for e in f["attrs"]])
        results.append(
            "    {} is at {:.2f} percent {}, maximum {:.2f} percent".format(
                attrs, f["current"], value_to_check, f["maximum"]
            )
        )

    result += "\n".join(results)

    return result


def get_check_from_overrides(overrides, default_check, groupings):
    """Get the overrides dict from overrides with the same groupings as groupings,
    or return the default"""
    checks = [o for o in overrides if o["groupings"] == groupings]
    if len(checks) == 0:
        return default_check
    elif len(checks) == 1:
        return checks[0]
    else:
        group_string = ", ".join([f"{k}: {v}" for k, v in groupings.items()])
        paasta_print("UNKNOWN Multiple overrides specified for %s" % group_string)
        sys.exit(3)


def read_overrides(override_file):
    if override_file:
        with open(override_file, "r") as f:
            return json.loads(f.read())
    else:
        return {}


def run_capacity_check():
    options = parse_capacity_check_options()
    system_paasta_config = load_system_paasta_config()
    cluster = (
        options.cluster
        if options.cluster is not None
        else system_paasta_config.get_cluster()
    )
    value_to_check = options.type

    client = get_paasta_api_client(cluster=cluster)
    if client is None:
        paasta_print("UNKNOWN Failed to load paasta api client")
        sys.exit(3)

    overrides = read_overrides(options.overrides)

    attributes = options.attributes.split(",")

    try:
        resource_use = client.resources.resources(groupings=attributes).result()
    except HTTPError as e:
        paasta_print("UNKNOWN received exception from paasta api:\n\t%s" % e)
        sys.exit(3)

    default_check = {
        "warn": {"cpus": options.warn, "mem": options.warn, "disk": options.warn},
        "crit": {"cpus": options.crit, "mem": options.crit, "disk": options.crit},
    }

    failures = defaultdict(list)
    for usage_value in resource_use:
        check = get_check_from_overrides(
            overrides, default_check, usage_value["groupings"]
        )
        usage_percent = calc_percent_usage(usage_value, value_to_check)
        for c in ["crit", "warn"]:
            if usage_percent > check[c][value_to_check]:
                failures[c].append(
                    {
                        "attrs": [
                            {"attr": a, "value": v}
                            for a, v in usage_value["groupings"].items()
                        ],
                        "maximum": check[c][value_to_check],
                        "current": usage_percent,
                    }
                )
                break

    return_value = [0]
    if len(failures["crit"]) > 0:
        result = error_message(failures["crit"], "CRITICAL", cluster, value_to_check)
        paasta_print(result)
        return_value.append(2)
    if len(failures["warn"]) > 0:
        result = error_message(failures["warn"], "WARNING", cluster, value_to_check)
        paasta_print(result)
        return_value.append(1)

    if max(return_value) == 0:
        paasta_print(
            f"OK cluster {cluster} is below critical capacity in {value_to_check}"
        )

    sys.exit(max(return_value))


if __name__ == "__main__":
    run_capacity_check()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Sequence
from typing import Tuple

from bravado.exception import HTTPError

from paasta_tools.api.client import get_paasta_api_client
from paasta_tools.cli.utils import get_paasta_metastatus_cmd_args
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import SystemPaastaConfig


def add_subparser(subparsers,) -> None:
    status_parser = subparsers.add_parser(
        "metastatus",
        help="Display the status for an entire PaaSTA cluster",
        description=(
            "'paasta metastatus' is used to get the vital statistics about a PaaSTA "
            "cluster as a whole. This tool is helpful when answering the question: 'Is "
            "it just my service or the whole cluster that is broken?'\n\n"
            "metastatus operates by ssh'ing to a Mesos master of a remote cluster, and "
            "querying the local APIs."
        ),
        epilog=(
            "The metastatus command may time out during heavy load. When that happens "
            "users may execute the ssh command directly, in order to bypass the timeout."
        ),
    )
    status_parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help="""Print out more output regarding the state of the cluster.
        Multiple v options increase verbosity. Maximum is 3.""",
    )
    clusters_help = (
        "A comma separated list of clusters to view. Defaults to view all clusters. "
        "Try: --clusters norcal-prod,nova-prod"
    )
    status_parser.add_argument(
        "-c", "--clusters", help=clusters_help
    ).completer = lazy_choices_completer(list_clusters)
    status_parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    status_parser.add_argument(
        "-a",
        "--autoscaling-info",
        action="store_true",
        default=False,
        dest="autoscaling_info",
        help="Show cluster autoscaling info, implies -vv",
    )
    status_parser.add_argument(
        "--use-mesos-cache",
        action="store_true",
        default=False,
        dest="use_mesos_cache",
        help="Use Mesos cache for state.json and frameworks",
    )
    status_parser.add_argument(
        "-g",
        "--groupings",
        nargs="+",
        default=["region"],
        help=(
            "Group resource information of slaves grouped by attribute."
            "Note: This is only effective with -vv"
        ),
    )
    # The service and instance args default to None if not specified.
    status_parser.add_argument(
        "-s",
        "--service",
        help=(
            "Show how many of a given service instance can be run on a cluster slave."
            "Note: This is only effective with -vvv and --instance must also be specified"
        ),
    ).completer = lazy_choices_completer(list_services)
    status_parser.add_argument(
        "-i",
        "--instance",
        help=(
            "Show how many of a given service instance can be run on a cluster slave."
            "Note: This is only effective with -vvv and --service must also be specified"
        ),
    )
    status_parser.set_defaults(command=paasta_metastatus)


def paasta_metastatus_on_api_endpoint(
    cluster: str,
    system_paasta_config: SystemPaastaConfig,
    groupings: Sequence[str],
    verbose: int,
    autoscaling_info: bool = False,
    use_mesos_cache: bool = False,
) -> Tuple[int, str]:
    client = get_paasta_api_client(cluster, system_paasta_config)
    if not client:
        paasta_print("Cannot get a paasta-api client")
        exit(1)

    try:
        cmd_args, _ = get_paasta_metastatus_cmd_args(
            groupings=groupings,
            verbose=verbose,
            autoscaling_info=autoscaling_info,
            use_mesos_cache=use_mesos_cache,
        )
        res = client.metastatus.metastatus(
            cmd_args=[str(arg) for arg in cmd_args]
        ).result()
        output, exit_code = res.output, res.exit_code
    except HTTPError as exc:
        output, exit_code = exc.response.text, exc.status_code

    return exit_code, output


def print_cluster_status(
    cluster: str,
    system_paasta_config: SystemPaastaConfig,
    groupings: Sequence[str],
    verbose: int = 0,
    autoscaling_info: bool = False,
    use_mesos_cache: bool = False,
) -> int:
    """With a given cluster and verboseness, returns the status of the cluster
    output is printed directly to provide dashboards even if the cluster is unavailable"""
    return_code, output = paasta_metastatus_on_api_endpoint(
        cluster=cluster,
        system_paasta_config=system_paasta_config,
        groupings=groupings,
        verbose=verbose,
        autoscaling_info=autoscaling_info,
        use_mesos_cache=use_mesos_cache,
    )

    paasta_print("Cluster: %s" % cluster)
    paasta_print(get_cluster_dashboards(cluster))
    paasta_print(output)
    paasta_print()

    return return_code


def figure_out_clusters_to_inspect(args, all_clusters) -> Sequence[str]:
    if args.clusters is not None:
        clusters_to_inspect = args.clusters.split(",")
    else:
        clusters_to_inspect = all_clusters
    return clusters_to_inspect


def get_cluster_dashboards(cluster: str,) -> str:
    """Returns the direct dashboards for humans to use for a given cluster"""
    SPACER = " "
    try:
        dashboards = load_system_paasta_config().get_dashboard_links()[cluster]
    except KeyError as e:
        if e.args[0] == cluster:
            output = [PaastaColors.red("No dashboards configured for %s!" % cluster)]
        else:
            output = [PaastaColors.red("No dashboards configured!")]
    else:
        output = ["Dashboards:"]
        spacing = max((len(label) for label in dashboards.keys())) + 1
        for label, urls in dashboards.items():
            if isinstance(urls, list):
                urls = "\n    %s" % "\n    ".join(urls)
            output.append(
                "  {}:{}{}".format(
                    label, SPACER * (spacing - len(label)), PaastaColors.cyan(urls)
                )
            )
    return "\n".join(output)


def paasta_metastatus(args,) -> int:
    """Print the status of a PaaSTA clusters"""
    soa_dir = args.soa_dir
    system_paasta_config = load_system_paasta_config()

    all_clusters = list_clusters(soa_dir=soa_dir)
    clusters_to_inspect = figure_out_clusters_to_inspect(args, all_clusters)
    return_codes = []
    for cluster in clusters_to_inspect:
        if cluster in all_clusters:
            return_codes.append(
                print_cluster_status(
                    cluster=cluster,
                    system_paasta_config=system_paasta_config,
                    groupings=args.groupings,
                    verbose=args.verbose,
                    autoscaling_info=args.autoscaling_info,
                    use_mesos_cache=args.use_mesos_cache,
                )
            )
        else:
            paasta_print(
                "Cluster %s doesn't look like a valid cluster?" % args.clusters
            )
            paasta_print("Try using tab completion to help complete the cluster name")
    return 0 if all([return_code == 0 for return_code in return_codes]) else 1



import logging
from typing import Any
from typing import Sequence

from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet

from paasta_tools.kubernetes.application.controller_wrappers import Application
from paasta_tools.kubernetes.application.controller_wrappers import DeploymentWrapper
from paasta_tools.kubernetes.application.controller_wrappers import StatefulSetWrapper
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import sanitise_kubernetes_name

log = logging.getLogger(__name__)


def is_valid_application(deployment: V1Deployment):
    is_valid = (
        "paasta.yelp.com/service" in deployment.metadata.labels
        and "paasta.yelp.com/instance" in deployment.metadata.labels
        and "paasta.yelp.com/git_sha" in deployment.metadata.labels
        and "paasta.yelp.com/config_sha" in deployment.metadata.labels
    )
    if not is_valid:
        log.warning(
            f"deployment/{deployment.metadata.name} in "
            f"namespace/{deployment.metadata.namespace} "
            "does not have complete set of labels"
        )
        log.warning(deployment)
    return is_valid


def list_namespaced_deployments(
    kube_client: KubeClient, namespace: str, **kwargs
) -> Sequence[DeploymentWrapper]:
    return [
        DeploymentWrapper(deployment)
        for deployment in kube_client.deployments.list_namespaced_deployment(
            namespace, **kwargs
        ).items
        if is_valid_application(deployment)
    ]


def list_namespaced_stateful_sets(
    kube_client: KubeClient, namespace: str, **kwargs
) -> Sequence[StatefulSetWrapper]:
    return [
        StatefulSetWrapper(deployment)
        for deployment in kube_client.deployments.list_namespaced_stateful_set(
            namespace, **kwargs
        ).items
        if is_valid_application(deployment)
    ]


def list_namespaced_applications(
    kube_client: KubeClient, namespace: str, application_types: Sequence[Any]
) -> Sequence[Application]:
    """
    List all applications in the namespace of the types from application_types.
    Only applications with complete set of labels are included (See is_valid_application()).
    :param kube_client:
    :param namespace:
    :param application_types:  types of applications
    :return:
    """
    apps = []  # type: ignore
    for application_type in application_types:
        if application_type == V1Deployment:
            apps.extend(list_namespaced_deployments(kube_client, namespace))
        elif application_type == V1StatefulSet:
            apps.extend(list_namespaced_stateful_sets(kube_client, namespace))
    return apps


def get_app_name(service: str, instance: str):
    return sanitise_kubernetes_name(f"{service}-{instance}")



#!/usr/bin/env python
# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import datetime
import sys
import time

import a_sync

from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.utils import load_system_paasta_config


OUTPUT_FORMAT = "{:<30}  {:<8}  {:<20}  {:<27}  {}"
FRAMEWORK_NAME = "marathon"
MAX_BOUNCE_TIME_IN_HOURS = 4


def parse_args():
    parser = argparse.ArgumentParser(
        description="Find all containers serving previous push versions."
    )
    parser.add_argument(
        "--bounce-time",
        dest="bounce_time",
        type=int,
        default=MAX_BOUNCE_TIME_IN_HOURS,
        help=(
            "Ignore versions that were launched in the last BOUNCE_TIME hours "
            "because they probably are still bouncing."
        ),
    )
    return parser.parse_args()


def get_mesos_state():
    state = a_sync.block(get_mesos_master(use_mesos_cache=True).state)
    return state


def marathon_tasks(state):
    for framework in state.get("frameworks", []):
        if framework["name"].lower().startswith(FRAMEWORK_NAME):
            for task in framework.get("tasks", []):
                yield task


def create_slave_id_to_hostname_dict(state):
    res = {}
    for slave in state["slaves"]:
        res[slave["id"]] = slave["hostname"]
    return res


def group_running_tasks_by_id_and_gitsha(state):
    res = {}
    for t in marathon_tasks(state):
        if t["state"] == "TASK_RUNNING":
            task_id = t["name"][: t["name"].find(".", t["name"].find(".") + 1)]
            gitsha = t["name"][len(task_id) + 1 : t["name"].find(".", len(task_id) + 1)]
            res.setdefault(task_id, {}).setdefault(gitsha, []).append(t)
    return res


def detect_outdated_gitshas(versions, max_bounce_time_in_hours):
    """Find versions that should have drained more than 'max_bounce_time_in_hours' ago"""
    if len(versions) < 2:
        return []
    deploy_time = {}
    latest_deploy = 0
    for version, tasks in versions.items():
        deploy_time[version] = sum(t["statuses"][0]["timestamp"] for t in tasks) / len(
            tasks
        )
        if (
            deploy_time[version] > latest_deploy
            and time.time() - deploy_time[version] > max_bounce_time_in_hours * 3600
        ):
            latest_deploy = deploy_time[version]
    return [version for version, dtime in deploy_time.items() if dtime < latest_deploy]


def report_outdated_instances(task_id, gitsha, tasks, slave_id2hostname):
    output = []
    remedy = []
    for t in tasks:
        deploy_time = datetime.datetime.fromtimestamp(
            int(t["statuses"][0]["timestamp"])
        ).strftime("%Y-%m-%d %H:%M:%S")
        container_name = "mesos-{}.{}".format(
            t["slave_id"], t["statuses"][0]["container_status"]["container_id"]["value"]
        )
        hostname = slave_id2hostname[t["slave_id"]]
        hostname = hostname[: hostname.find(".")]
        service_instance = task_id.replace("--", "_")
        output.append(
            OUTPUT_FORMAT.format(
                service_instance[:30], gitsha[3:], deploy_time, hostname, container_name
            )
        )
        remedy.append(
            'ssh {0} "sudo hadown {1}; sleep 10; sudo docker stop {2}; sudo haup {1}"'.format(
                hostname, service_instance, container_name
            )
        )
    return output, remedy


def check_mesos_tasks(max_bounce_time_in_hours=MAX_BOUNCE_TIME_IN_HOURS):
    output = []
    remedy = []
    state = get_mesos_state()
    aggregated_tasks = group_running_tasks_by_id_and_gitsha(state)
    slave_id2hostname = create_slave_id_to_hostname_dict(state)
    for task_id, versions in aggregated_tasks.items():
        for gitsha in detect_outdated_gitshas(versions, max_bounce_time_in_hours):
            temp_output, temp_remedy = report_outdated_instances(
                task_id, gitsha, versions[gitsha], slave_id2hostname
            )
            output.extend(temp_output)
            remedy.extend(temp_remedy)
    return output, remedy


def main():
    args = parse_args()
    cluster = load_system_paasta_config().get_cluster()
    output, remedy = check_mesos_tasks(args.bounce_time)
    if output:
        print(
            "CRITICAL - There are {} tasks running in {} that are more than {}h older than their"
            " last bounce.".format(len(output), cluster, args.bounce_time)
        )
        print(
            OUTPUT_FORMAT.format(
                "SERVICE.INSTANCE", "COMMIT", "CREATED", "HOSTNAME", "CONTAINER"
            )
        )
        print("\n".join(output))
        print("")
        print("Run the following commands to terminate them:")
        print("{code}")
        print("\n".join(remedy))
        print("{code}")
        return 1
    else:
        print(f"OK - There are no outdated tasks in {cluster}")
        return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./check_flink_services_health.py [options]
"""
import datetime
import logging
from typing import Sequence
from typing import Tuple

import pysensu_yelp

from paasta_tools import flink_tools
from paasta_tools.check_services_replication_tools import main
from paasta_tools.flink_tools import FlinkDeploymentConfig
from paasta_tools.kubernetes_tools import filter_pods_by_service_instance
from paasta_tools.kubernetes_tools import is_pod_ready
from paasta_tools.kubernetes_tools import V1Pod
from paasta_tools.monitoring_tools import check_under_replication
from paasta_tools.monitoring_tools import send_replication_event
from paasta_tools.smartstack_tools import KubeSmartstackReplicationChecker
from paasta_tools.utils import is_under_replicated


log = logging.getLogger(__name__)


def container_lifetime(pod: V1Pod,) -> datetime.timedelta:
    """Return a time duration for how long the pod is alive
    """
    st = pod.status.start_time
    return datetime.datetime.now(st.tzinfo) - st


def healthy_flink_containers_cnt(si_pods: Sequence[V1Pod], container_type: str) -> int:
    """Return count of healthy Flink containers with given type
    """
    return len(
        [
            pod
            for pod in si_pods
            if pod.metadata.labels["flink-container-type"] == container_type
            and is_pod_ready(pod)
            and container_lifetime(pod).total_seconds() > 60
        ]
    )


def check_under_registered_taskmanagers(
    instance_config: FlinkDeploymentConfig, expected_count: int,
) -> Tuple[bool, str]:
    """Check if not enough taskmanagers have been registered to the jobmanager and
    returns both the result of the check in the form of a boolean and a human-readable
    text to be used in logging or monitoring events.
    """
    unhealthy = True
    try:
        overview = flink_tools.get_flink_jobmanager_overview(
            instance_config.service, instance_config.instance, instance_config.cluster
        )
        num_reported = overview.get("taskmanagers", 0)
        crit_threshold = instance_config.get_replication_crit_percentage()
        output = (
            f"Service {instance_config.job_id} has "
            f"{num_reported} out of {expected_count} expected instances "
            f"of taskmanager reported by dashboard!\n"
            f"(threshold: {crit_threshold}%)"
        )
        unhealthy, _ = is_under_replicated(num_reported, expected_count, crit_threshold)
    except ValueError as e:
        output = (
            f"Dashboard of service {instance_config.job_id} is not available!\n({e})"
        )
    if unhealthy:
        output += f"""
What this alert means:

  This alert means that the Flink dashboard is not reporting the expected
  number of taskmanagers.

Reasons this might be happening:

  The service may simply be unhealthy. There also may not be enough resources
  in the cluster to support the requested instance count.

Things you can do:

  * Fix the cause of the unhealthy service. Try running:

     paasta status -s {instance_config.service} -i {instance_config.instance} -c {instance_config.cluster} -vv

"""
    return (unhealthy, output)


def check_flink_service_health(
    instance_config: FlinkDeploymentConfig,
    all_pods: Sequence[V1Pod],
    smartstack_replication_checker: KubeSmartstackReplicationChecker,
) -> None:
    si_pods = filter_pods_by_service_instance(
        pod_list=all_pods,
        service=instance_config.service,
        instance=instance_config.instance,
    )
    taskmanagers_expected_cnt = instance_config.config_dict.get(
        "taskmanager", {"instances": 10}
    ).get("instances", 10)
    num_healthy_supervisors = healthy_flink_containers_cnt(si_pods, "supervisor")
    num_healthy_jobmanagers = healthy_flink_containers_cnt(si_pods, "jobmanager")
    num_healthy_taskmanagers = healthy_flink_containers_cnt(si_pods, "taskmanager")

    results = [
        check_under_replication(
            instance_config=instance_config,
            expected_count=1,
            num_available=num_healthy_supervisors,
            sub_component="supervisor",
        ),
        check_under_replication(
            instance_config=instance_config,
            expected_count=1,
            num_available=num_healthy_jobmanagers,
            sub_component="jobmanager",
        ),
        check_under_replication(
            instance_config=instance_config,
            expected_count=taskmanagers_expected_cnt,
            num_available=num_healthy_taskmanagers,
            sub_component="taskmanager",
        ),
        check_under_registered_taskmanagers(
            instance_config=instance_config, expected_count=taskmanagers_expected_cnt,
        ),
    ]
    output = "\n########\n".join([r[1] for r in results])
    if any(r[0] for r in results):
        log.error(output)
        status = pysensu_yelp.Status.CRITICAL
    else:
        log.info(output)
        status = pysensu_yelp.Status.OK
    send_replication_event(
        instance_config=instance_config, status=status, output=output
    )


if __name__ == "__main__":
    main(
        flink_tools.FlinkDeploymentConfig,
        check_flink_service_health,
        namespace="paasta-flinks",
    )



"""iptables helper functions.

Unlike the `firewall` module, these functions know nothing about PaaSTA and
could effectively be a third-party library. They just make working with
iptables a little bit easier.
"""
import collections
import contextlib
import logging

import iptc

log = logging.getLogger(__name__)


RULE_TARGET_SORT_ORDER = {
    # all else defaults to '0'
    "LOG": 1,
    "REJECT": 2.0,
}


_RuleBase = collections.namedtuple(
    "Rule", ("protocol", "src", "dst", "target", "matches", "target_parameters")
)


class Rule(_RuleBase):
    """Rule representation.

    Working with iptc's rule classes directly doesn't work well, since rules
    represent actual existing iptables rules, and changes are applied
    immediately. They're also difficult to compare.
    """

    def __new__(cls, *args, **kwargs):
        result = _RuleBase.__new__(cls, *args, **kwargs)
        result.validate()
        return result

    def _replace(self, **kwargs):
        result = super()._replace(**kwargs)
        result.validate()
        return result

    def validate(self):
        if self.target == "REJECT":
            assert any(
                name == "reject-with" for name, _ in self.target_parameters
            ), "REJECT rules must specify reject-with"
        assert tuple(sorted(self.matches)) == self.matches, "matches should be sorted"
        for match_name, params in self.matches:
            for param_name, param_value in params:
                assert (
                    "_" not in param_name
                ), f"use dashes instead of underscores in {param_name}"
                assert isinstance(
                    param_value, tuple
                ), f"value of {param_name} should be tuple"
        assert (
            tuple(sorted(self.target_parameters)) == self.target_parameters
        ), "target_parameters should be sorted"
        for param_name, param_value in self.target_parameters:
            assert (
                "_" not in param_name
            ), f"use dashes instead of underscores in {param_name}"
            assert isinstance(
                param_value, tuple
            ), f"value of {param_name} should be tuple"

    @classmethod
    def from_iptc(cls, rule):
        fields = {
            "protocol": rule.protocol,
            "src": rule.src,
            "dst": rule.dst,
            "target": rule.target.name,
            "matches": (),
            "target_parameters": (),
        }

        for param_name, param_value in sorted(rule.target.get_all_parameters().items()):
            fields["target_parameters"] += ((param_name, tuple(param_value)),)

        matches = []
        for match in rule.matches:
            matches.append(
                (
                    match.name,
                    tuple(
                        (param, tuple(value))
                        for param, value in sorted(match.get_all_parameters().items())
                    ),
                )
            )
        # ensure that matches are sorted for consistency with matching
        fields["matches"] = tuple(sorted(matches))

        return cls(**fields)

    def to_iptc(self):
        rule = iptc.Rule()
        rule.protocol = self.protocol
        rule.src = self.src
        rule.dst = self.dst
        target = rule.create_target(self.target)
        for param_name, param_value in self.target_parameters:
            target.set_parameter(param_name, param_value)
        for name, params in self.matches:
            match = rule.create_match(name)
            for param_name, param_value in params:
                match.set_parameter(param_name, param_value)
        return rule


@contextlib.contextmanager
def iptables_txn(table):
    """Temporarily disable autocommit and commit at the end.

    If an exception occurs, changes are rolled back.

    By default, changes to iptables rules are applied immediately. In some
    cases, we want to avoid that.

    https://github.com/ldx/python-iptables#autocommit
    """
    assert table.autocommit is True, table.autocommit
    try:
        table.autocommit = False
        yield
        table.commit()
    finally:
        table.refresh()
        table.autocommit = True


class ChainDoesNotExist(Exception):
    pass


def all_chains():
    return {chain.name for chain in iptc.Table(iptc.Table.FILTER).chains}


def ensure_chain(chain, rules):
    """Idempotently ensure a chain exists and has an exact set of rules.

    This function creates or updates an existing chain to match the rules
    passed in.

    This function will not reorder existing rules, but any new rules are always
    inserted at the front of the chain.
    """
    try:
        current_rules = set(list_chain(chain))
    except ChainDoesNotExist:
        create_chain(chain)
        current_rules = set()

    for rule in rules:
        if rule not in current_rules:
            insert_rule(chain, rule)

    extra_rules = current_rules - set(rules)
    if extra_rules:
        delete_rules(chain, extra_rules)


def _rule_sort_key(rule_tuple):
    old_index, rule = rule_tuple
    target_name = rule.target
    return (RULE_TARGET_SORT_ORDER.get(target_name, 0), old_index)


def reorder_chain(chain_name):
    """Ensure that any REJECT rules are last, and any LOG rules are second-to-last
    """

    table = iptc.Table(iptc.Table.FILTER)
    with iptables_txn(table):
        rules = list_chain(chain_name)
        chain = iptc.Chain(table, chain_name)

        # sort the rules by rule_key, which uses (RULE_TARGET_SORT_ORDER, idx)
        sorted_rules_with_indices = sorted(enumerate(rules), key=_rule_sort_key)

        for new_index, (old_index, rule) in enumerate(sorted_rules_with_indices):
            if new_index == old_index:
                continue
            log.debug(f"reordering chain {chain_name} rule {rule} to #{new_index}")
            chain.replace_rule(rule.to_iptc(), new_index)


def ensure_rule(chain, rule):
    rules = list_chain(chain)
    if rule not in rules:
        insert_rule(chain, rule)


def insert_rule(chain_name, rule):
    log.debug(f"adding rule to {chain_name}: {rule}")
    chain = iptc.Chain(iptc.Table(iptc.Table.FILTER), chain_name)
    chain.insert_rule(rule.to_iptc())


def delete_rules(chain_name, rules):
    log.debug(f"deleting rules from {chain_name}: {rules}")
    table = iptc.Table(iptc.Table.FILTER)
    with iptables_txn(table):
        chain = iptc.Chain(table, chain_name)
        for potential_rule in chain.rules:
            if Rule.from_iptc(potential_rule) in rules:
                chain.delete_rule(potential_rule)


def create_chain(chain_name):
    log.debug(f"creating chain: {chain_name}")
    iptc.Table(iptc.Table.FILTER).create_chain(chain_name)


def delete_chain(chain_name):
    log.debug(f"deleting chain: {chain_name}")
    chain = iptc.Chain(iptc.Table(iptc.Table.FILTER), chain_name)
    chain.flush()
    chain.delete()


def list_chain(chain_name):
    """List rules in a chain.

    Returns a list of iptables rules, or raises ChainDoesNotExist.
    """
    table = iptc.Table(iptc.Table.FILTER)
    chain = iptc.Chain(table, chain_name)
    # TODO: is there any way to do this without listing all chains? (probably slow)
    # If the chain doesn't exist, chain.rules will be an empty list, so we need
    # to make sure the chain actually _does_ exist.
    if chain in table.chains:
        return tuple(Rule.from_iptc(rule) for rule in chain.rules)
    else:
        raise ChainDoesNotExist(chain_name)



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_job.py <service.instance> [options]

Command line options:

- -d <SOA_DIR>, --soa-dir <SOA_DIR>: Specify a SOA config dir to read from
- -v, --verbose: Verbose output
"""
import argparse
import logging
import sys
from typing import Optional
from typing import Sequence
from typing import Tuple

from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet

from paasta_tools.kubernetes.application.controller_wrappers import Application
from paasta_tools.kubernetes.application.controller_wrappers import DeploymentWrapper
from paasta_tools.kubernetes.application.controller_wrappers import StatefulSetWrapper
from paasta_tools.kubernetes_tools import ensure_namespace
from paasta_tools.kubernetes_tools import InvalidKubernetesConfig
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import list_all_deployments
from paasta_tools.kubernetes_tools import load_kubernetes_service_config_no_cache
from paasta_tools.utils import decompose_job_id
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InvalidJobNameError
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import NoDeploymentsAvailable
from paasta_tools.utils import SPACER

log = logging.getLogger(__name__)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates marathon jobs.")
    parser.add_argument(
        "service_instance_list",
        nargs="+",
        help="The list of marathon service instances to create or update",
        metavar="SERVICE%sINSTANCE" % SPACER,
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "-c", "--cluster", dest="cluster", help="paasta cluster",
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    args = parser.parse_args()
    return args


def main() -> None:
    args = parse_args()
    soa_dir = args.soa_dir
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    # system_paasta_config = load_system_paasta_config()
    kube_client = KubeClient()

    ensure_namespace(kube_client, namespace="paasta")
    setup_kube_succeeded = setup_kube_deployments(
        kube_client=kube_client,
        service_instances=args.service_instance_list,
        soa_dir=soa_dir,
        cluster=args.cluster or load_system_paasta_config().get_cluster(),
    )
    sys.exit(0 if setup_kube_succeeded else 1)


def validate_job_name(service_instance: str) -> bool:
    try:
        service, instance, _, __ = decompose_job_id(service_instance)
    except InvalidJobNameError:
        log.error(
            "Invalid service instance specified. Format is service%sinstance." % SPACER
        )
        return False
    return True


def setup_kube_deployments(
    kube_client: KubeClient,
    service_instances: Sequence[str],
    cluster: str,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> bool:
    if service_instances:
        existing_kube_deployments = set(list_all_deployments(kube_client))
        existing_apps = {
            (deployment.service, deployment.instance)
            for deployment in existing_kube_deployments
        }
    service_instances_with_valid_names = [
        decompose_job_id(service_instance)
        for service_instance in service_instances
        if validate_job_name(service_instance)
    ]
    applications = [
        create_application_object(
            kube_client=kube_client,
            service=service_instance[0],
            instance=service_instance[1],
            cluster=cluster,
            soa_dir=soa_dir,
        )
        for service_instance in service_instances_with_valid_names
    ]

    for _, app in applications:
        if (
            app
            and (app.kube_deployment.service, app.kube_deployment.instance)
            not in existing_apps
        ):
            app.create(kube_client)
        elif app and app.kube_deployment not in existing_kube_deployments:
            app.update(kube_client)
        else:
            log.debug(f"{app} is up to date, no action taken")

    return (False, None) not in applications and len(
        service_instances_with_valid_names
    ) == len(service_instances)


def create_application_object(
    kube_client: KubeClient, service: str, instance: str, cluster: str, soa_dir: str,
) -> Tuple[bool, Optional[Application]]:
    try:
        service_instance_config = load_kubernetes_service_config_no_cache(
            service, instance, cluster, soa_dir=soa_dir,
        )
    except NoDeploymentsAvailable:
        log.debug(
            "No deployments found for %s.%s in cluster %s. Skipping."
            % (service, instance, cluster)
        )
        return True, None
    except NoConfigurationForServiceError:
        error_msg = (
            "Could not read kubernetes configuration file for %s.%s in cluster %s"
            % (service, instance, cluster)
        )
        log.error(error_msg)
        return False, None

    try:
        formatted_application = service_instance_config.format_kubernetes_app()
    except InvalidKubernetesConfig as e:
        log.error(str(e))
        return False, None

    app: Optional[Application] = None
    if isinstance(formatted_application, V1Deployment):
        app = DeploymentWrapper(formatted_application)
    elif isinstance(formatted_application, V1StatefulSet):
        app = StatefulSetWrapper(formatted_application)
    else:
        raise Exception("Unknown kubernetes object to update")

    app.load_local_config(soa_dir, cluster)
    return True, app


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import contextlib
import shutil
import sys

from cookiecutter.main import cookiecutter

from paasta_tools.cli.fsm.autosuggest import suggest_smartstack_proxy_port
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors


@contextlib.contextmanager
def make_copyfile_symlink_aware():
    """The reasoning behind this monkeypatch is that cookiecutter doesn't
    respect symlinks at all, and at Yelp we use symlinks to reduce duplication
    in the soa configs. Maybe cookie-cutter will accept a symlink-aware PR?
    """
    orig_copyfile = shutil.copyfile
    orig_copymode = shutil.copymode

    def symlink_aware_copyfile(*args, **kwargs):
        kwargs.setdefault("follow_symlinks", False)
        orig_copyfile(*args, **kwargs)

    def symlink_aware_copymode(*args, **kwargs):
        kwargs.setdefault("follow_symlinks", False)
        orig_copymode(*args, **kwargs)

    shutil.copyfile = symlink_aware_copyfile
    shutil.copymode = symlink_aware_copymode
    try:
        yield
    finally:
        shutil.copyfile = orig_copyfile
        shutil.copymode = orig_copymode


def parse_args():
    fsm_parser = argparse.ArgumentParser(
        "fsm",
        description=(
            "'paasta fsm' is used to generate example soa-configs, which is useful during initial "
            "service creation. Currently 'fsm' generates 'yelp-specific' configuration, but can still "
            "be used as an example of a fully working PaaSTA service.\n\n"
            "After 'paasta fsm' is run, the operator should inspect the generated boilerplate configuration "
            "and adjust it to meet the particular needs of the new service."
        ),
    )
    fsm_parser.add_argument(
        "-y",
        "--yelpsoa-config-root",
        dest="yelpsoa_config_root",
        default=".",
        help=(
            "Path to root of yelpsoa-configs checkout\n"
            "Defaults to current working directory"
        ),
    )
    fsm_parser.set_defaults(command=paasta_fsm)
    return fsm_parser.parse_args()


def get_paasta_config(yelpsoa_config_root):
    variables = {"proxy_port": suggest_smartstack_proxy_port(yelpsoa_config_root)}
    return variables


def write_paasta_config(variables, template, destination):
    paasta_print("Using cookiecutter template from %s" % template)
    with make_copyfile_symlink_aware():
        cookiecutter(
            template=template,
            extra_context=variables,
            output_dir=destination,
            overwrite_if_exists=True,
            no_input=not sys.stdout.isatty(),
        )


def paasta_fsm(args):
    variables = get_paasta_config(yelpsoa_config_root=args.yelpsoa_config_root)
    destination = args.yelpsoa_config_root

    paasta_config = load_system_paasta_config()
    template = paasta_config.get_fsm_template()

    write_paasta_config(variables=variables, template=template, destination=destination)

    paasta_print(PaastaColors.yellow("               _  _(o)_(o)_  _"))
    paasta_print(PaastaColors.red(r"             ._\`:_ F S M _:' \_,"))
    paasta_print(PaastaColors.green(r"                 / (`---'\ `-."))
    paasta_print(PaastaColors.cyan("              ,-`  _)    (_,"))
    paasta_print("With My Noodly Appendage I Have Written Configs!")
    paasta_print()
    paasta_print("Customize Them If It Makes You Happy -- http://y/paasta For Details")
    paasta_print("Remember To Add, Commit, And Push When You're Done:")
    paasta_print()


def main():
    args = parse_args()
    paasta_fsm(args)


if __name__ == "__main__":
    main()



import logging
import threading
from abc import ABC
from abc import abstractmethod
from time import sleep
from typing import Optional
from typing import Union

from kubernetes.client import V1beta1PodDisruptionBudget
from kubernetes.client import V1DeleteOptions
from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet
from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import create_deployment
from paasta_tools.kubernetes_tools import create_pod_disruption_budget
from paasta_tools.kubernetes_tools import create_stateful_set
from paasta_tools.kubernetes_tools import force_delete_pods
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import KubeDeployment
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import list_all_deployments
from paasta_tools.kubernetes_tools import load_kubernetes_service_config_no_cache
from paasta_tools.kubernetes_tools import pod_disruption_budget_for_service_instance
from paasta_tools.kubernetes_tools import update_deployment
from paasta_tools.kubernetes_tools import update_stateful_set


class Application(ABC):
    def __init__(
        self,
        item: Union[V1Deployment, V1StatefulSet],
        logging=logging.getLogger(__name__),
    ) -> None:
        """
        This Application wrapper is an interface for creating/deleting k8s deployments and statefulsets
        soa_config is KubernetesDeploymentConfig. It is not loaded in init because it is not always required.
        :param item: Kubernetes Object(V1Deployment/V1StatefulSet) that has already been filled up.
        :param logging: where logs go
        """
        if not item.metadata.namespace:
            item.metadata.namespace = "paasta"
        self.kube_deployment = KubeDeployment(
            service=item.metadata.labels["paasta.yelp.com/service"],
            instance=item.metadata.labels["paasta.yelp.com/instance"],
            git_sha=item.metadata.labels["paasta.yelp.com/git_sha"],
            config_sha=item.metadata.labels["paasta.yelp.com/config_sha"],
            replicas=item.spec.replicas,
        )
        self.item = item
        self.soa_config = None  # type: KubernetesDeploymentConfig
        self.logging = logging

    def load_local_config(
        self, soa_dir: str, cluster: str
    ) -> Optional[KubernetesDeploymentConfig]:
        if not self.soa_config:
            self.soa_config = load_kubernetes_service_config_no_cache(
                service=self.kube_deployment.service,
                instance=self.kube_deployment.instance,
                cluster=cluster,
                soa_dir=soa_dir,
            )
        return self.soa_config

    def __str__(self):
        service = self.kube_deployment.service
        instance = self.kube_deployment.instance
        git_sha = self.kube_deployment.git_sha
        config_sha = self.kube_deployment.config_sha
        return f"{service}-{instance}-{git_sha}-{config_sha}"

    @abstractmethod
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        pass

    def create(self, kube_client: KubeClient):
        """
        Create all controllers, HPA, and pod disruption budgets related to this application
        :param kube_client:
        """
        pass

    def update(self, kube_client: KubeClient):
        """
        Update all controllers, HPA, and pod disruption budgets related to this application
        :param kube_client:
        """
        pass

    def delete_pod_disruption_budget(self, kube_client: KubeClient) -> None:
        try:
            kube_client.policy.delete_namespaced_pod_disruption_budget(
                name=self.item.metadata.name,
                namespace=self.item.metadata.namespace,
                body=V1DeleteOptions(),
            )
        except ApiException as e:
            if e.status == 404:
                # Deployment does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent pod disruption budget/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise
        else:
            self.logging.info(
                "deleted pod disruption budget/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )

    def ensure_pod_disruption_budget(
        self, kube_client: KubeClient
    ) -> V1beta1PodDisruptionBudget:
        pdr = pod_disruption_budget_for_service_instance(
            service=self.kube_deployment.service,
            instance=self.kube_deployment.instance,
            max_unavailable="{}%".format(
                int((1 - self.soa_config.get_bounce_margin_factor()) * 100)
            ),
        )
        try:
            existing_pdr = kube_client.policy.read_namespaced_pod_disruption_budget(
                name=pdr.metadata.name, namespace=pdr.metadata.namespace
            )
        except ApiException as e:
            if e.status == 404:
                existing_pdr = None
            else:
                raise

        if existing_pdr:
            if existing_pdr.spec.min_available is not None:
                logging.debug(
                    "Not updating poddisruptionbudget: can't have both "
                    "min_available and max_unavailable"
                )
            elif existing_pdr.spec.max_unavailable != pdr.spec.max_unavailable:
                logging.debug(f"Updating poddisruptionbudget {pdr.metadata.name}")
                return kube_client.policy.patch_namespaced_pod_disruption_budget(
                    name=pdr.metadata.name, namespace=pdr.metadata.namespace, body=pdr
                )
            else:
                logging.debug(f"poddisruptionbudget {pdr.metadata.name} up to date")
        else:
            logging.debug(f"creating poddisruptionbudget {pdr.metadata.name}")
            return create_pod_disruption_budget(
                kube_client=kube_client, pod_disruption_budget=pdr
            )


class DeploymentWrapper(Application):
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        delete_options = V1DeleteOptions(propagation_policy="Foreground")
        try:
            kube_client.deployments.delete_namespaced_deployment(
                self.item.metadata.name,
                self.item.metadata.namespace,
                body=delete_options,
            )
        except ApiException as e:
            if e.status == 404:
                # Deployment does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent deploy/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise
        else:
            self.logging.info(
                "deleted deploy/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )
        self.delete_pod_disruption_budget(kube_client)
        self.delete_horizontal_pod_autoscaler(kube_client)

    def get_existing_app(self, kube_client: KubeClient):
        return kube_client.deployments.read_namespaced_deployment(
            name=self.item.metadata.name, namespace=self.item.metadata.namespace
        )

    def create(self, kube_client: KubeClient) -> None:
        create_deployment(kube_client=kube_client, formatted_deployment=self.item)
        self.ensure_pod_disruption_budget(kube_client)
        self.sync_horizontal_pod_autoscaler(kube_client)

    def deep_delete_and_create(self, kube_client: KubeClient) -> None:
        self.deep_delete(kube_client)
        timer = 0
        while (
            self.kube_deployment in set(list_all_deployments(kube_client))
            and timer < 60
        ):
            sleep(1)
            timer += 1

        if timer >= 60 and self.kube_deployment in set(
            list_all_deployments(kube_client)
        ):
            try:
                force_delete_pods(
                    self.item.metadata.name,
                    self.kube_deployment.service,
                    self.kube_deployment.instance,
                    self.item.metadata.namespace,
                    kube_client,
                )
            except ApiException as e:
                if e.status == 404:
                    # Deployment does not exist, nothing to delete but
                    # we can consider this a success.
                    self.logging.debug(
                        "not deleting nonexistent deploy/{} from namespace/{}".format(
                            self.kube_deployment.service, self.item.metadata.namespace
                        )
                    )
                else:
                    raise
        else:
            self.logging.info(
                "deleted deploy/{} from namespace/{}".format(
                    self.kube_deployment.service, self.item.metadata.namespace
                )
            )
        self.create(kube_client=kube_client)

    def update(self, kube_client: KubeClient) -> None:
        # If HPA is enabled, do not update replicas.
        # In all other cases, replica is set to max(instances, min_instances)
        if self.soa_config.config_dict.get("bounce_method", "") == "brutal":
            threading.Thread(
                target=self.deep_delete_and_create, args=[KubeClient()]
            ).start()
            return
        if self.should_have_hpa() is None:
            self.item.spec.replicas = self.get_existing_app(kube_client).spec.replicas
        update_deployment(kube_client=kube_client, formatted_deployment=self.item)
        self.ensure_pod_disruption_budget(kube_client)
        self.sync_horizontal_pod_autoscaler(kube_client)

    def should_have_hpa(self):
        return (
            self.soa_config.get_max_instances() is not None
            # with bespoke autoscaler, setup_kubernetes_job sets the number of instances directly; no HPA is required.
            and self.soa_config.get_autoscaling_params()["decision_policy"] != "bespoke"
        )

    def sync_horizontal_pod_autoscaler(self, kube_client: KubeClient) -> None:
        """
        In order for autoscaling to work, there needs to be at least two configurations
        min_instnace, max_instance, and there cannot be instance.
        """
        self.logging.info(
            f"Syncing HPA setting for {self.item.metadata.name}/name in {self.item.metadata.namespace}"
        )
        hpa_exists = self.exists_hpa(kube_client)
        # NO autoscaling
        if not self.should_have_hpa():
            # Remove HPA if autoscaling is disabled
            if hpa_exists:
                self.delete_horizontal_pod_autoscaler(kube_client)
            return

        body = self.soa_config.get_autoscaling_metric_spec(
            name=self.item.metadata.name,
            cluster=self.soa_config.cluster,
            namespace=self.item.metadata.namespace,
        )
        if not body:
            raise Exception(
                f"CRIT: autoscaling misconfigured for {self.kube_deployment.service}."
                + f"{self.kube_deployment.instance}.Please correct the configuration and update pre-commit hook."
            )
        self.logging.debug(body)
        if not hpa_exists:
            self.logging.info(
                f"Creating new HPA for {self.item.metadata.name}/name in {self.item.metadata.namespace}"
            )
            kube_client.autoscaling.create_namespaced_horizontal_pod_autoscaler(
                namespace=self.item.metadata.namespace, body=body, pretty=True
            )
        else:
            self.logging.info(
                f"Updating new HPA for {self.item.metadata.name}/name in {self.item.metadata.namespace}/namespace"
            )
            kube_client.autoscaling.patch_namespaced_horizontal_pod_autoscaler(
                name=self.item.metadata.name,
                namespace=self.item.metadata.namespace,
                body=body,
                pretty=True,
            )

    def exists_hpa(self, kube_client: KubeClient) -> bool:
        return (
            len(
                kube_client.autoscaling.list_namespaced_horizontal_pod_autoscaler(
                    field_selector=f"metadata.name={self.item.metadata.name}",
                    namespace=self.item.metadata.namespace,
                ).items
            )
            > 0
        )

    def delete_horizontal_pod_autoscaler(self, kube_client: KubeClient) -> None:
        try:
            kube_client.autoscaling.delete_namespaced_horizontal_pod_autoscaler(
                name=self.item.metadata.name,
                namespace=self.item.metadata.namespace,
                body=V1DeleteOptions(),
            )
        except ApiException as e:
            if e.status == 404:
                # Deployment does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    f"not deleting nonexistent HPA/{self.item.metadata.name} from namespace/{self.item.metadata.namespace}"
                )
            else:
                raise
        else:
            self.logging.info(
                "deleted HPA/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )


class StatefulSetWrapper(Application):
    def deep_delete(self, kube_client: KubeClient) -> None:
        """
        Remove all controllers, pods, and pod disruption budgets related to this application
        :param kube_client:
        """
        delete_options = V1DeleteOptions(propagation_policy="Foreground")
        try:
            kube_client.deployments.delete_namespaced_stateful_set(
                self.item.metadata.name,
                self.item.metadata.namespace,
                body=delete_options,
            )
        except ApiException as e:
            if e.status == 404:
                # StatefulSet does not exist, nothing to delete but
                # we can consider this a success.
                self.logging.debug(
                    "not deleting nonexistent statefulset/{} from namespace/{}".format(
                        self.item.metadata.name, self.item.metadata.namespace
                    )
                )
            else:
                raise
        else:
            self.logging.info(
                "deleted statefulset/{} from namespace/{}".format(
                    self.item.metadata.name, self.item.metadata.namespace
                )
            )
        self.delete_pod_disruption_budget(kube_client)

    def create(self, kube_client: KubeClient):
        create_stateful_set(kube_client=kube_client, formatted_stateful_set=self.item)
        self.ensure_pod_disruption_budget(kube_client)

    def update(self, kube_client: KubeClient):
        update_stateful_set(kube_client=kube_client, formatted_stateful_set=self.item)
        self.ensure_pod_disruption_budget(kube_client)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import itertools

from . import exceptions
from paasta_tools.utils import paasta_print


async def get_files_for_tasks(task_list, file_list, max_workers):
    no_files_found = True

    async def process(task_fname):
        task, fname = task_fname
        try:
            fobj = await task.file(fname)
        except exceptions.SlaveDoesNotExist:
            if task is None:
                paasta_print(f"(Unknown Task):{fname} (Slave no longer exists)")
            else:
                paasta_print(f"{task['id']}:{task_fname} (Slave no longer exists)")
            raise exceptions.SkipResult

        if await fobj.exists():
            return fobj

    elements = itertools.chain(
        *[[(task, fname) for fname in file_list] for task in task_list]
    )

    futures = [asyncio.ensure_future(process(element)) for element in elements]

    if futures:
        for result in asyncio.as_completed(futures):
            try:
                result = await result
                if result:
                    no_files_found = False
                    yield result
            except exceptions.SkipResult:
                pass

    if no_files_found:
        raise exceptions.FileNotFoundForTaskException(
            "None of the tasks in {} contain the files in list {}".format(
                ",".join([task["id"] for task in task_list]), ",".join(file_list)
            )
        )



import datetime
import time
import traceback
from typing import NamedTuple

import humanize

from paasta_tools import marathon_tools
from paasta_tools.deployd.common import BounceTimers
from paasta_tools.deployd.common import DelayDeadlineQueueProtocol
from paasta_tools.deployd.common import exponential_back_off
from paasta_tools.deployd.common import PaastaThread
from paasta_tools.deployd.common import ServiceInstance
from paasta_tools.metrics.metrics_lib import BaseMetrics
from paasta_tools.setup_marathon_job import deploy_marathon_service
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import SystemPaastaConfig


class BounceResults(NamedTuple):
    bounce_again_in_seconds: float
    return_code: int


class PaastaDeployWorker(PaastaThread):
    def __init__(
        self,
        worker_number: int,
        instances_to_bounce: DelayDeadlineQueueProtocol,
        config: SystemPaastaConfig,
        metrics_provider: BaseMetrics,
    ) -> None:
        super().__init__()
        self.daemon = True
        self.name = f"Worker{worker_number}"
        self.instances_to_bounce = instances_to_bounce
        self.metrics = metrics_provider
        self.config = config
        self.cluster = self.config.get_cluster()
        self.busy = False
        self.setup()

    def setup(self) -> None:
        system_paasta_config = load_system_paasta_config()
        self.marathon_servers = marathon_tools.get_marathon_servers(
            system_paasta_config
        )
        self.marathon_clients = marathon_tools.get_marathon_clients(
            self.marathon_servers
        )

    def setup_timers(self, service_instance: ServiceInstance) -> BounceTimers:
        bounce_length_timer = self.metrics.create_timer(
            "bounce_length_timer",
            service=service_instance.service,
            instance=service_instance.instance,
            paasta_cluster=self.cluster,
        )

        processed_by_worker_timer = self.metrics.create_timer(
            "processed_by_worker",
            service=service_instance.service,
            instance=service_instance.instance,
            paasta_cluster=self.cluster,
        )
        setup_marathon_timer = self.metrics.create_timer(
            "setup_marathon_timer",
            service=service_instance.service,
            instance=service_instance.instance,
            paasta_cluster=self.cluster,
        )
        return BounceTimers(
            processed_by_worker=processed_by_worker_timer,
            setup_marathon=setup_marathon_timer,
            bounce_length=bounce_length_timer,
        )

    def run(self) -> None:
        """Takes things from the to_bounce_now queue, processes them, then
        might put them on the bounce_later queue for future processing"""
        self.log.info(f"{self.name} starting up")
        while True:
            with self.instances_to_bounce.get() as service_instance:
                self.busy = True
                try:
                    (
                        bounce_again_in_seconds,
                        return_code,
                    ) = self.process_service_instance(service_instance)
                except Exception:
                    self.log.error(
                        f"{self.name} Worker failed to process service instance and will retry. "
                        f"Caused by exception: {traceback.format_exc()}"
                    )
                    return_code = -2
                failures = service_instance.failures
                if return_code != 0:
                    failures = service_instance.failures + 1
                    bounce_again_in_seconds = exponential_back_off(
                        failures=failures,
                        factor=self.config.get_deployd_worker_failure_backoff_factor(),
                        base=2,
                        max_time=6000,
                    )
                if bounce_again_in_seconds:
                    bounce_by = int(time.time()) + bounce_again_in_seconds
                    service_instance = ServiceInstance(
                        service=service_instance.service,
                        instance=service_instance.instance,
                        bounce_by=bounce_by,
                        wait_until=bounce_by,
                        watcher=self.name,
                        failures=failures,
                        processed_count=service_instance.processed_count + 1,
                        bounce_start_time=service_instance.bounce_start_time,
                        enqueue_time=time.time(),
                    )
                    self.instances_to_bounce.put(service_instance)
            self.busy = False
            time.sleep(0.1)

    def process_service_instance(
        self, service_instance: ServiceInstance
    ) -> BounceResults:
        bounce_timers = self.setup_timers(service_instance)
        if service_instance.enqueue_time is not None:
            bounce_timers.processed_by_worker.record(
                time.time() - service_instance.enqueue_time
            )

        human_bounce_by = humanize.naturaldelta(
            datetime.timedelta(seconds=(time.time() - service_instance.bounce_by))
        )
        self.log.info(
            f"{self.name} processing {service_instance.service}.{service_instance.instance} (bounce_by {human_bounce_by} ago)"
        )  # noqa E501

        bounce_timers.setup_marathon.start()
        return_code, bounce_again_in_seconds = deploy_marathon_service(
            service=service_instance.service,
            instance=service_instance.instance,
            clients=self.marathon_clients,
            soa_dir=marathon_tools.DEFAULT_SOA_DIR,
            marathon_apps_with_clients=None,
        )

        bounce_timers.setup_marathon.stop()
        self.log.info(
            f"{self.name} setup marathon completed with exit code {return_code} for {service_instance.service}.{service_instance.instance}"
        )  # noqa E501
        if bounce_again_in_seconds:
            self.log.info(
                f"{self.name} {service_instance.service}.{service_instance.instance} not in steady state so bouncing again in {bounce_again_in_seconds}"
            )  # noqa E501
        else:
            self.log.info(
                f"{self.name} {service_instance.service}.{service_instance.instance} in steady state"
            )
            if service_instance.processed_count > 0:
                bounce_timers.bounce_length.record(
                    time.time() - service_instance.bounce_start_time
                )
        return BounceResults(bounce_again_in_seconds, return_code)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import io
import itertools
import logging
import sys
from typing import Mapping
from typing import MutableSequence
from typing import Optional
from typing import Sequence
from typing import Tuple

import a_sync
from marathon.exceptions import MarathonError
from mypy_extensions import TypedDict

from paasta_tools import __version__
from paasta_tools.autoscaling.autoscaling_cluster_lib import AutoscalingInfo
from paasta_tools.autoscaling.autoscaling_cluster_lib import (
    get_autoscaling_info_for_all_resources,
)
from paasta_tools.cli.utils import get_instance_config
from paasta_tools.kubernetes_tools import is_kubernetes_available
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.marathon_tools import get_marathon_clients
from paasta_tools.marathon_tools import get_marathon_servers
from paasta_tools.marathon_tools import MarathonClient
from paasta_tools.marathon_tools import MarathonClients
from paasta_tools.mesos.exceptions import MasterNotAvailableException
from paasta_tools.mesos.master import MesosMaster
from paasta_tools.mesos.master import MesosState
from paasta_tools.mesos_tools import get_mesos_leader
from paasta_tools.mesos_tools import get_mesos_master
from paasta_tools.mesos_tools import is_mesos_available
from paasta_tools.metrics import metastatus_lib
from paasta_tools.metrics.metastatus_lib import _GenericNodeGroupingFunctionT
from paasta_tools.metrics.metastatus_lib import _KeyFuncRetT
from paasta_tools.metrics.metastatus_lib import HealthCheckResult
from paasta_tools.metrics.metastatus_lib import ResourceUtilization
from paasta_tools.metrics.metastatus_lib import ResourceUtilizationDict
from paasta_tools.utils import format_table
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import paasta_print
from paasta_tools.utils import PaastaColors
from paasta_tools.utils import print_with_indent
from paasta_tools.utils import set_paasta_print_file


log = logging.getLogger("paasta_metastatus")
logging.basicConfig()
# kazoo can be really noisy - turn it down
logging.getLogger("kazoo").setLevel(logging.CRITICAL)
logging.getLogger("paasta_tools.autoscaling.autoscaling_cluster_lib").setLevel(
    logging.ERROR
)

ServiceInstanceStats = TypedDict(
    "ServiceInstanceStats", {"mem": float, "cpus": float, "disk": float, "gpus": int}
)


class FatalError(Exception):
    def __init__(self, exit_code: int) -> None:
        self.exit_code = exit_code


def parse_args(argv):
    parser = argparse.ArgumentParser(description="")
    parser.add_argument(
        "-g",
        "--groupings",
        nargs="+",
        default=["pool", "region"],
        help=(
            "Group resource information of slaves grouped by attribute."
            "Note: This is only effective with -vv"
        ),
    )
    parser.add_argument("-t", "--threshold", type=int, default=90)
    parser.add_argument("--use-mesos-cache", action="store_true", default=False)
    parser.add_argument(
        "-a",
        "--autoscaling-info",
        action="store_true",
        default=False,
        dest="autoscaling_info",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        dest="verbose",
        default=0,
        help="Print out more output regarding the state of the cluster",
    )
    parser.add_argument(
        "-s",
        "--service",
        help=(
            "Show how many of a given service instance can be run on a cluster slave."
            "Note: This is only effective with -vvv and --instance must also be specified"
        ),
    )
    parser.add_argument(
        "-i",
        "--instance",
        help=(
            "Show how many of a given service instance can be run on a cluster slave."
            "Note: This is only effective with -vvv and --service must also be specified"
        ),
    )
    return parser.parse_args(argv)


def get_marathon_framework_ids(
    marathon_clients: Sequence[MarathonClient],
) -> Sequence[str]:
    return [client.get_info().framework_id for client in marathon_clients]


def _run_mesos_checks(
    mesos_master: MesosMaster, mesos_state: MesosState
) -> Sequence[HealthCheckResult]:
    mesos_state_status = metastatus_lib.get_mesos_state_status(mesos_state)

    metrics = a_sync.block(mesos_master.metrics_snapshot)
    mesos_metrics_status = metastatus_lib.get_mesos_resource_utilization_health(
        mesos_metrics=metrics, mesos_state=mesos_state
    )
    return mesos_state_status + mesos_metrics_status  # type: ignore


def _run_marathon_checks(
    marathon_clients: Sequence[MarathonClient],
) -> Sequence[HealthCheckResult]:
    try:
        marathon_results = metastatus_lib.get_marathon_status(marathon_clients)
        return marathon_results
    except (MarathonError, ValueError) as e:
        paasta_print(
            PaastaColors.red(f"CRITICAL: Unable to contact Marathon cluster: {e}")
        )
        raise FatalError(2)


def all_marathon_clients(
    marathon_clients: MarathonClients,
) -> Sequence[MarathonClient]:
    return [
        c for c in itertools.chain(marathon_clients.current, marathon_clients.previous)
    ]


def utilization_table_by_grouping(
    groupings: Sequence[str],
    grouping_function: _GenericNodeGroupingFunctionT,
    resource_info_dict_grouped: Mapping[_KeyFuncRetT, ResourceUtilizationDict],
    threshold: float,
    service_instance_stats: Optional[ServiceInstanceStats] = None,
) -> Tuple[Sequence[MutableSequence[str]], bool]:
    static_headers = [
        "CPU (used/total)",
        "RAM (used/total)",
        "Disk (used/total)",
        "GPU (used/total)",
        "Agent count",
    ]
    # service_instance_stats could be None so check and insert a header if needed.
    if service_instance_stats:
        # Insert so agent count is still last
        static_headers.insert(-1, "Slots + Limiting Resource")

    all_rows = [[grouping.capitalize() for grouping in groupings] + static_headers]
    table_rows = []

    for grouping_values, resource_info_dict in resource_info_dict_grouped.items():
        resource_utilizations = metastatus_lib.resource_utillizations_from_resource_info(
            total=resource_info_dict["total"], free=resource_info_dict["free"]
        )
        healthcheck_utilization_pairs = [
            metastatus_lib.healthcheck_result_resource_utilization_pair_for_resource_utilization(
                utilization, threshold
            )
            for utilization in resource_utilizations
        ]
        healthy_exit = all(pair[0].healthy for pair in healthcheck_utilization_pairs)
        table_rows.append(
            metastatus_lib.get_table_rows_for_resource_info_dict(
                [v for g, v in grouping_values], healthcheck_utilization_pairs
            )
        )
        # Fill table rows with service-instance data if possible.
        if service_instance_stats:
            fill_table_rows_with_service_instance_stats(
                service_instance_stats, resource_utilizations, table_rows
            )

        # Always append the agent count last
        table_rows[-1].append(str(resource_info_dict["slave_count"]))

    table_rows = sorted(table_rows, key=lambda x: x[0 : len(groupings)])
    all_rows.extend(table_rows)

    return all_rows, healthy_exit


def utilization_table_by_grouping_from_mesos_state(
    groupings: Sequence[str],
    threshold: float,
    mesos_state: MesosState,
    service_instance_stats: Optional[ServiceInstanceStats] = None,
) -> Tuple[Sequence[MutableSequence[str]], bool]:
    grouping_function = metastatus_lib.key_func_for_attribute_multi(groupings)
    resource_info_dict_grouped = metastatus_lib.get_resource_utilization_by_grouping(
        grouping_function, mesos_state
    )

    return utilization_table_by_grouping(
        groupings,
        grouping_function,
        resource_info_dict_grouped,
        threshold,
        service_instance_stats,
    )


def utilization_table_by_grouping_from_kube(
    groupings: Sequence[str],
    threshold: float,
    kube_client: KubeClient,
    service_instance_stats: Optional[ServiceInstanceStats] = None,
) -> Tuple[Sequence[MutableSequence[str]], bool]:
    grouping_function = metastatus_lib.key_func_for_attribute_multi_kube(groupings)
    resource_info_dict_grouped = metastatus_lib.get_resource_utilization_by_grouping_kube(
        grouping_function, kube_client
    )

    return utilization_table_by_grouping(
        groupings,
        grouping_function,
        resource_info_dict_grouped,
        threshold,
        service_instance_stats,
    )


def fill_table_rows_with_service_instance_stats(
    service_instance_stats: ServiceInstanceStats,
    resource_utilizations: Sequence[ResourceUtilization],
    table_rows: MutableSequence[MutableSequence[str]],
) -> None:
    # Calculate the max number of runnable service instances given the current resources (e.g. cpus, mem, disk)
    resource_free_dict = {rsrc.metric: rsrc.free for rsrc in resource_utilizations}
    num_service_instances_allowed = float("inf")
    limiting_factor = "Unknown"
    # service_instance_stats.keys() should be a subset of resource_free_dict
    for rsrc_name, rsrc_amt_wanted in service_instance_stats.items():
        if rsrc_amt_wanted > 0:  # type: ignore
            # default=0 to indicate there is none of that resource
            rsrc_free = resource_free_dict.get(rsrc_name, 0)
            if (
                rsrc_free // rsrc_amt_wanted  # type: ignore
                < num_service_instances_allowed  # type: ignore
            ):
                limiting_factor = rsrc_name
                num_service_instances_allowed = (
                    rsrc_free // rsrc_amt_wanted  # type: ignore
                )
    table_rows[-1].append(
        "{:6} ; {}".format(int(num_service_instances_allowed), limiting_factor)
    )


def get_service_instance_stats(
    service: str, instance: str, cluster: str
) -> Optional[ServiceInstanceStats]:
    """Returns a Dict with stats about a given service instance.

    Args:
        service: the service name
        instance: the instance name
        cluster: the cluster name where the service instance will be searched for

    Returns:
        A Dict mapping resource name to the amount of that resource the particular service instance consumes.
    """
    if service is None or instance is None or cluster is None:
        return None

    try:
        instance_config = get_instance_config(service, instance, cluster)
        # Get all fields that are showed in the 'paasta metastatus -vvv' command
        if instance_config.get_gpus():
            gpus = int(instance_config.get_gpus())
        else:
            gpus = 0
        service_instance_stats = ServiceInstanceStats(
            mem=instance_config.get_mem(),
            cpus=instance_config.get_cpus(),
            disk=instance_config.get_disk(),
            gpus=gpus,
        )
        return service_instance_stats
    except Exception as e:
        log.error(
            f"Failed to get stats for service {service} instance {instance}: {str(e)}"
        )
        return None


def _run_kube_checks(kube_client: KubeClient,) -> Sequence[HealthCheckResult]:
    kube_status = metastatus_lib.get_kube_status(kube_client)
    kube_metrics_status = metastatus_lib.get_kube_resource_utilization_health(
        kube_client=kube_client
    )
    return kube_status + kube_metrics_status  # type: ignore


def print_output(argv: Optional[Sequence[str]] = None) -> None:
    mesos_available = is_mesos_available()
    kube_available = is_kubernetes_available()

    args = parse_args(argv)

    system_paasta_config = load_system_paasta_config()

    if mesos_available:
        master_kwargs = {}
        # we don't want to be passing False to not override a possible True
        # value from system config
        if args.use_mesos_cache:
            master_kwargs["use_mesos_cache"] = True

        master = get_mesos_master(**master_kwargs)

        marathon_servers = get_marathon_servers(system_paasta_config)
        marathon_clients = all_marathon_clients(get_marathon_clients(marathon_servers))

        try:
            mesos_state = a_sync.block(master.state)
            all_mesos_results = _run_mesos_checks(
                mesos_master=master, mesos_state=mesos_state
            )
        except MasterNotAvailableException as e:
            # if we can't connect to master at all,
            # then bomb out early
            paasta_print(PaastaColors.red("CRITICAL:  %s" % "\n".join(e.args)))
            raise FatalError(2)

        marathon_results = _run_marathon_checks(marathon_clients)
    else:
        marathon_results = [
            metastatus_lib.HealthCheckResult(
                message="Marathon is not configured to run here", healthy=True
            )
        ]
        all_mesos_results = [
            metastatus_lib.HealthCheckResult(
                message="Mesos is not configured to run here", healthy=True
            )
        ]

    if kube_available:
        kube_client = KubeClient()
        kube_results = _run_kube_checks(kube_client)
    else:
        kube_results = [
            metastatus_lib.HealthCheckResult(
                message="Kubernetes is not configured to run here", healthy=True
            )
        ]

    mesos_ok = all(metastatus_lib.status_for_results(all_mesos_results))
    marathon_ok = all(metastatus_lib.status_for_results(marathon_results))
    kube_ok = all(metastatus_lib.status_for_results(kube_results))

    mesos_summary = metastatus_lib.generate_summary_for_check("Mesos", mesos_ok)
    marathon_summary = metastatus_lib.generate_summary_for_check(
        "Marathon", marathon_ok
    )
    kube_summary = metastatus_lib.generate_summary_for_check("Kubernetes", kube_ok)

    healthy_exit = True if all([mesos_ok, marathon_ok]) else False

    paasta_print(f"Master paasta_tools version: {__version__}")
    paasta_print("Mesos leader: %s" % get_mesos_leader())
    metastatus_lib.print_results_for_healthchecks(
        mesos_summary, mesos_ok, all_mesos_results, args.verbose
    )
    if args.verbose > 1 and mesos_available:
        print_with_indent("Resources Grouped by %s" % ", ".join(args.groupings), 2)
        all_rows, healthy_exit = utilization_table_by_grouping_from_mesos_state(
            groupings=args.groupings, threshold=args.threshold, mesos_state=mesos_state
        )
        for line in format_table(all_rows):
            print_with_indent(line, 4)

        if args.autoscaling_info:
            print_with_indent("Autoscaling resources:", 2)
            headers = [
                field.replace("_", " ").capitalize()
                for field in AutoscalingInfo._fields
            ]
            table = [headers] + [
                [str(x) for x in asi]
                for asi in get_autoscaling_info_for_all_resources(mesos_state)
            ]

            for line in format_table(table):
                print_with_indent(line, 4)

        if args.verbose >= 3:
            print_with_indent("Per Slave Utilization", 2)
            cluster = system_paasta_config.get_cluster()
            service_instance_stats = get_service_instance_stats(
                args.service, args.instance, cluster
            )
            if service_instance_stats:
                print_with_indent(
                    "Service-Instance stats:" + str(service_instance_stats), 2
                )
            # print info about slaves here. Note that we don't make modifications to
            # the healthy_exit variable here, because we don't care about a single slave
            # having high usage.
            all_rows, _ = utilization_table_by_grouping_from_mesos_state(
                groupings=args.groupings + ["hostname"],
                threshold=args.threshold,
                mesos_state=mesos_state,
                service_instance_stats=service_instance_stats,
            )
            # The last column from utilization_table_by_grouping_from_mesos_state is "Agent count", which will always be
            # 1 for per-slave resources, so delete it.
            for row in all_rows:
                row.pop()

            for line in format_table(all_rows):
                print_with_indent(line, 4)
    metastatus_lib.print_results_for_healthchecks(
        marathon_summary, marathon_ok, marathon_results, args.verbose
    )
    metastatus_lib.print_results_for_healthchecks(
        kube_summary, kube_ok, kube_results, args.verbose
    )
    if args.verbose > 1 and kube_available:
        print_with_indent("Resources Grouped by %s" % ", ".join(args.groupings), 2)
        all_rows, healthy_exit = utilization_table_by_grouping_from_kube(
            groupings=args.groupings, threshold=args.threshold, kube_client=kube_client
        )
        for line in format_table(all_rows):
            print_with_indent(line, 4)

        if args.autoscaling_info:
            print_with_indent("No autoscaling resources for Kubernetes", 2)

        if args.verbose >= 3:
            print_with_indent("Per Node Utilization", 2)
            cluster = system_paasta_config.get_cluster()
            service_instance_stats = get_service_instance_stats(
                args.service, args.instance, cluster
            )
            if service_instance_stats:
                print_with_indent(
                    "Service-Instance stats:" + str(service_instance_stats), 2
                )
            # print info about nodes here. Note that we don't make
            # modifications to the healthy_exit variable here, because we don't
            # care about a single node having high usage.
            all_rows, _ = utilization_table_by_grouping_from_kube(
                groupings=args.groupings + ["hostname"],
                threshold=args.threshold,
                kube_client=kube_client,
                service_instance_stats=service_instance_stats,
            )
            # The last column from utilization_table_by_grouping_from_kube is "Agent count", which will always be
            # 1 for per-node resources, so delete it.
            for row in all_rows:
                row.pop()

            for line in format_table(all_rows):
                print_with_indent(line, 4)

    if not healthy_exit:
        raise FatalError(2)


def get_output(argv: Optional[Sequence[str]] = None) -> Tuple[str, int]:
    output = io.StringIO()
    exit_code = 1
    with set_paasta_print_file(output):
        exit_code = 0
        try:
            print_output(argv)
        except FatalError as e:
            exit_code = e.exit_code
    ret = output.getvalue()
    return ret, exit_code


def main(argv: Optional[Sequence[str]] = None) -> None:
    exit_code = 0
    try:
        print_output(argv)
    except FatalError as e:
        exit_code = e.exit_code
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import inspect
import logging
import socket
import sys
import time
from queue import Empty
from typing import Any
from typing import List
from typing import Type

import service_configuration_lib
from kazoo.client import KazooClient

from paasta_tools.deployd import watchers
from paasta_tools.deployd.common import DelayDeadlineQueue
from paasta_tools.deployd.common import DelayDeadlineQueueProtocol
from paasta_tools.deployd.common import get_marathon_clients_from_config
from paasta_tools.deployd.common import PaastaQueue
from paasta_tools.deployd.common import PaastaThread
from paasta_tools.deployd.common import ServiceInstance
from paasta_tools.deployd.leader import PaastaLeaderElection
from paasta_tools.deployd.metrics import QueueAndWorkerMetrics
from paasta_tools.deployd.queue import ZKDelayDeadlineQueue
from paasta_tools.deployd.workers import PaastaDeployWorker
from paasta_tools.list_marathon_service_instances import (
    get_service_instances_that_need_bouncing,
)
from paasta_tools.marathon_tools import DEFAULT_SOA_DIR
from paasta_tools.metrics.metrics_lib import get_metrics_interface
from paasta_tools.utils import get_services_for_cluster
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import ZookeeperPool

# Broken out into a constant so that we don't get drift between this and the code in paasta_deployd_steps.py that
# searches for this message.
DEAD_DEPLOYD_WORKER_MESSAGE = "Detected a dead worker, starting a replacement thread"


class AddHostnameFilter(logging.Filter):
    def __init__(self) -> None:
        super().__init__()
        self.hostname = socket.gethostname()

    def filter(self, record: Any) -> bool:
        record.hostname = self.hostname
        return True


class DeployDaemon(PaastaThread):
    def __init__(self) -> None:
        super().__init__()
        self.started = False
        self.daemon = True
        service_configuration_lib.disable_yaml_cache()
        self.config = load_system_paasta_config()
        self.setup_logging()
        self.metrics = get_metrics_interface("paasta.deployd")
        self.setup_instances_to_bounce()
        self.control = PaastaQueue("ControlQueue")
        self.marathon_clients = get_marathon_clients_from_config()

    def setup_instances_to_bounce(self) -> None:
        if self.config.get_deployd_use_zk_queue():
            zk_client = KazooClient(hosts=self.config.get_zk_hosts())
            zk_client.start()
            self.instances_to_bounce: DelayDeadlineQueueProtocol = ZKDelayDeadlineQueue(
                client=zk_client
            )
        else:
            self.instances_to_bounce = DelayDeadlineQueue()

    def setup_logging(self) -> None:
        root_logger = logging.getLogger()
        root_logger.setLevel(getattr(logging, self.config.get_deployd_log_level()))
        handler = logging.StreamHandler()
        handler.addFilter(AddHostnameFilter())
        root_logger.addHandler(handler)
        logging.getLogger("kazoo").setLevel(logging.CRITICAL)
        handler.setFormatter(
            logging.Formatter(
                "%(asctime)s:%(hostname)s:%(levelname)s:%(name)s:%(message)s"
            )
        )

    def run(self) -> None:
        self.log.info("paasta-deployd starting up...")
        startup_counter = self.metrics.create_counter(
            "process_started", paasta_cluster=self.config.get_cluster()
        )
        startup_counter.count()
        with ZookeeperPool() as self.zk:
            self.election = PaastaLeaderElection(
                self.zk,
                "/paasta-deployd-leader",
                socket.getfqdn(),
                control=self.control,
            )
            self.is_leader = False
            self.log.info("Waiting to become leader")
            self.election.run(self.startup)
            self.log.info("Leadership given up, exiting...")

    @property
    def watcher_threads_enabled(self) -> List[Type[watchers.PaastaWatcher]]:
        disabled_watchers = self.config.get_disabled_watchers()
        watcher_classes = [
            obj[1]
            for obj in inspect.getmembers(watchers)
            if inspect.isclass(obj[1]) and obj[1].__bases__[0] == watchers.PaastaWatcher
        ]
        enabled_watchers = [
            x for x in watcher_classes if x.__name__ not in disabled_watchers
        ]
        return enabled_watchers

    def startup(self) -> None:
        self.is_leader = True
        self.log.info("This node is elected as leader {}".format(socket.getfqdn()))
        leader_counter = self.metrics.create_counter(
            "leader_elections", paasta_cluster=self.config.get_cluster()
        )
        leader_counter.count()
        self.log.info("Starting all watcher threads")
        self.start_watchers()
        self.log.info(
            "All watchers started, now adding all services for initial bounce"
        )
        self.add_all_services()
        self.log.info("Prioritising services that we know need a bounce...")
        if self.config.get_deployd_startup_oracle_enabled():
            self.prioritise_bouncing_services()
        self.log.info("Starting worker threads")
        self.start_workers()
        QueueAndWorkerMetrics(
            queue=self.instances_to_bounce,
            workers=self.workers,
            cluster=self.config.get_cluster(),
            metrics_provider=self.metrics,
        ).start()
        self.started = True
        self.log.info("Startup finished!")
        self.main_loop()

    def main_loop(self) -> None:
        while True:
            try:
                message = self.control.get(block=False)
            except Empty:
                message = None
            if message == "ABORT":
                self.log.info("Got ABORT message, main_loop exiting...")
                break
            if not self.all_watchers_running():
                self.log.error("One or more watcher died, committing suicide!")
                sys.exit(1)
            if self.all_workers_dead():
                self.log.error("All workers have died, committing suicide!")
                sys.exit(1)
            self.check_and_start_workers()
            time.sleep(0.1)

    def all_watchers_running(self) -> bool:
        return all([watcher.is_alive() for watcher in self.watcher_threads])

    def all_workers_dead(self) -> bool:
        return all([not worker.is_alive() for worker in self.workers])

    def check_and_start_workers(self) -> None:
        live_workers = len([worker for worker in self.workers if worker.is_alive()])
        number_of_dead_workers = self.config.get_deployd_number_workers() - live_workers
        for i in range(number_of_dead_workers):
            self.log.error(DEAD_DEPLOYD_WORKER_MESSAGE)
            worker_no = len(self.workers) + 1
            worker = PaastaDeployWorker(
                worker_no, self.instances_to_bounce, self.config, self.metrics
            )
            worker.start()
            self.workers.append(worker)

    def stop(self) -> None:
        self.control.put("ABORT")

    def start_workers(self) -> None:
        self.workers: List[PaastaDeployWorker] = []
        for i in range(self.config.get_deployd_number_workers()):
            worker = PaastaDeployWorker(
                i, self.instances_to_bounce, self.config, self.metrics
            )
            worker.start()
            self.workers.append(worker)

    def add_all_services(self) -> None:
        instances = get_services_for_cluster(
            cluster=self.config.get_cluster(),
            instance_type="marathon",
            soa_dir=DEFAULT_SOA_DIR,
        )
        for service, instance in instances:
            self.instances_to_bounce.put(
                ServiceInstance(
                    service=service,
                    instance=instance,
                    watcher="daemon_start",
                    bounce_by=time.time()
                    + self.config.get_deployd_startup_bounce_deadline(),
                    wait_until=time.time(),
                    failures=0,
                    bounce_start_time=time.time(),
                    enqueue_time=time.time(),
                )
            )

    def prioritise_bouncing_services(self) -> None:
        service_instances = get_service_instances_that_need_bouncing(
            self.marathon_clients, DEFAULT_SOA_DIR
        )

        now = time.time()

        for service_instance in service_instances:
            self.log.info(f"Prioritising {service_instance} to be bounced immediately")
            service, instance = service_instance.split(".")
            self.instances_to_bounce.put(
                ServiceInstance(
                    service=service,
                    instance=instance,
                    watcher=type(self).__name__,
                    bounce_by=now,
                    wait_until=now,
                    failures=0,
                    bounce_start_time=time.time(),
                    enqueue_time=time.time(),
                )
            )

    def start_watchers(self) -> None:
        """ should block until all threads happy"""
        self.watcher_threads = [
            watcher(
                instances_to_bounce=self.instances_to_bounce,
                cluster=self.config.get_cluster(),
                zookeeper_client=self.zk,
                config=self.config,
            )
            for watcher in self.watcher_threads_enabled
        ]

        self.log.info(f"Starting the following watchers {self.watcher_threads}")
        for watcher in self.watcher_threads:
            watcher.start()
        self.log.info("Waiting for all watchers to start")
        attempts = 0
        while attempts < 120:
            if all([watcher.is_ready for watcher in self.watcher_threads]):
                return
            self.log.info("Sleeping and waiting for watchers to all start")
            self.log.info(
                "Waiting on: {}".format(
                    [
                        watcher.__class__.__name__
                        for watcher in self.watcher_threads
                        if not watcher.is_ready
                    ]
                )
            )
            time.sleep(1)
            attempts += 1
        self.log.error("Failed to start all the watchers, exiting...")
        sys.exit(1)


def main() -> None:
    dd = DeployDaemon()
    dd.start()
    while dd.is_alive():
        time.sleep(0.1)


if __name__ == "__main__":
    main()



import os
import time
from datetime import datetime

import pytz
from tzlocal import get_localzone as tzlocal_get_localzone

from paasta_tools.api import client
from paasta_tools.utils import paasta_print


def get_localzone():
    if "TZ" in os.environ:
        return pytz.timezone(os.environ["TZ"])
    else:
        return tzlocal_get_localzone()


def get_service_autoscale_pause_time(cluster):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print(
            "Could not connect to paasta api. Maybe you misspelled the cluster?"
        )
        return 1
    pause_time, http = api.service_autoscaler.get_service_autoscaler_pause().result()
    if http.status_code == 500:
        paasta_print("Could not connect to zookeeper server")
        return 2

    pause_time = float(pause_time)
    if pause_time < time.time():
        paasta_print("Service autoscaler is not paused")
    else:
        local_tz = get_localzone()
        paused_readable = local_tz.localize(
            datetime.fromtimestamp(pause_time)
        ).strftime("%F %H:%M:%S %Z")
        paasta_print(f"Service autoscaler is paused until {paused_readable}")

    return 0


def update_service_autoscale_pause_time(cluster, mins):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print(
            "Could not connect to paasta api. Maybe you misspelled the cluster?"
        )
        return 1
    body = {"minutes": mins}
    res, http = api.service_autoscaler.update_service_autoscaler_pause(
        json_body=body
    ).result()
    if http.status_code == 500:
        paasta_print("Could not connect to zookeeper server")
        return 2

    paasta_print(f"Service autoscaler is paused for {mins}")
    return 0


def delete_service_autoscale_pause_time(cluster):
    api = client.get_paasta_api_client(cluster=cluster, http_res=True)
    if not api:
        paasta_print(
            "Could not connect to paasta api. Maybe you misspelled the cluster?"
        )
        return 1
    res, http = api.service_autoscaler.delete_service_autoscaler_pause().result()
    if http.status_code == 500:
        paasta_print("Could not connect to zookeeper server")
        return 2

    paasta_print("Service autoscaler is unpaused")
    return 0
